/*
 * GStreamer
 * Copyright (C) 2008 Filippo Argiolas <filippo.argiolas@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstglshader.h"

#define GST_GL_SHADER_GET_PRIVATE(o)					\
  (G_TYPE_INSTANCE_GET_PRIVATE((o), GST_GL_TYPE_SHADER, GstGLShaderPrivate))

enum
{
  PROP_0,
  PROP_VERTEX_SRC,
  PROP_FRAGMENT_SRC,
  PROP_COMPILED,
  PROP_ACTIVE                   /* unused */
};

struct _GstGLShaderPrivate
{
  gchar *vertex_src;
  gchar *fragment_src;

  GLhandleARB vertex_handle;
  GLhandleARB fragment_handle;
  GLhandleARB program_handle;

  gboolean compiled;
  gboolean active;
};

GST_DEBUG_CATEGORY_STATIC (gst_gl_shader_debug);
#define GST_CAT_DEFAULT gst_gl_shader_debug

#define DEBUG_INIT \
  GST_DEBUG_CATEGORY_INIT (gst_gl_shader_debug, "glshader", 0, "shader");
G_DEFINE_TYPE_WITH_CODE (GstGLShader, gst_gl_shader, G_TYPE_OBJECT, DEBUG_INIT);

static void
gst_gl_shader_finalize (GObject * object)
{
  GstGLShader *shader;
  GstGLShaderPrivate *priv;
  /*  GLint status = GL_FALSE; */
  /* GLenum err = 0; */

  shader = GST_GL_SHADER (object);
  priv = shader->priv;

  GST_TRACE ("finalizing shader %u", priv->program_handle);

  g_free (priv->vertex_src);
  g_free (priv->fragment_src);

  /* release shader objects */
  gst_gl_shader_release (shader);

  /* delete program */
  if (priv->program_handle) {
    GST_TRACE ("finalizing program shader %u", priv->program_handle);

    glDeleteObjectARB (priv->program_handle);
    /* err = glGetError (); */
    /* GST_WARNING ("error: 0x%x", err);  */
    /* glGetObjectParameterivARB(priv->program_handle, GL_OBJECT_DELETE_STATUS_ARB, &status); */
    /* GST_INFO ("program deletion status:%s", status == GL_TRUE ? "true" : "false" ); */
  }

  GST_DEBUG ("shader deleted %u", priv->program_handle);

  priv->fragment_handle = 0;
  priv->vertex_handle = 0;
  priv->program_handle = 0;

  G_OBJECT_CLASS (gst_gl_shader_parent_class)->finalize (object);
}

static void
gst_gl_shader_set_property (GObject * object,
    guint prop_id, const GValue * value, GParamSpec * pspec)
{
  GstGLShader *shader = GST_GL_SHADER (object);

  switch (prop_id) {
    case PROP_VERTEX_SRC:
      gst_gl_shader_set_vertex_source (shader, g_value_get_string (value));
      break;
    case PROP_FRAGMENT_SRC:
      gst_gl_shader_set_fragment_source (shader, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_gl_shader_get_property (GObject * object,
    guint prop_id, GValue * value, GParamSpec * pspec)
{
  GstGLShader *shader = GST_GL_SHADER (object);
  GstGLShaderPrivate *priv = shader->priv;

  switch (prop_id) {
    case PROP_VERTEX_SRC:
      g_value_set_string (value, priv->vertex_src);
      break;
    case PROP_FRAGMENT_SRC:
      g_value_set_string (value, priv->fragment_src);
      break;
    case PROP_COMPILED:
      g_value_set_boolean (value, priv->compiled);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }

}

static void
gst_gl_shader_class_init (GstGLShaderClass * klass)
{
  /* bind class methods .. */
  GObjectClass *obj_class = G_OBJECT_CLASS (klass);

  g_type_class_add_private (klass, sizeof (GstGLShaderPrivate));

  obj_class->finalize = gst_gl_shader_finalize;
  obj_class->set_property = gst_gl_shader_set_property;
  obj_class->get_property = gst_gl_shader_get_property;

  /* .. and install properties */

  g_object_class_install_property (obj_class,
      PROP_VERTEX_SRC,
      g_param_spec_string ("vertex-src",
          "Vertex Source",
          "GLSL Vertex Shader source code", NULL,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  g_object_class_install_property (obj_class,
      PROP_FRAGMENT_SRC,
      g_param_spec_string ("fragment-src",
          "Fragment Source",
          "GLSL Fragment Shader source code", NULL,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  g_object_class_install_property (obj_class,
      PROP_ACTIVE,
      g_param_spec_string ("active",
          "Active", "Enable/Disable the shader", NULL,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
  g_object_class_install_property (obj_class,
      PROP_COMPILED,
      g_param_spec_boolean ("compiled",
          "Compiled",
          "Shader compile and link status", FALSE,
          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
}

void
gst_gl_shader_set_vertex_source (GstGLShader * shader, const gchar * src)
{
  GstGLShaderPrivate *priv;

  g_return_if_fail (GST_GL_IS_SHADER (shader));
  g_return_if_fail (src != NULL);

  priv = shader->priv;

  if (gst_gl_shader_is_compiled (shader))
    gst_gl_shader_release (shader);

  g_free (priv->vertex_src);

  priv->vertex_src = g_strdup (src);
}

void
gst_gl_shader_set_fragment_source (GstGLShader * shader, const gchar * src)
{
  GstGLShaderPrivate *priv;

  g_return_if_fail (GST_GL_IS_SHADER (shader));
  g_return_if_fail (src != NULL);

  priv = shader->priv;

  if (gst_gl_shader_is_compiled (shader))
    gst_gl_shader_release (shader);

  g_free (priv->fragment_src);

  priv->fragment_src = g_strdup (src);
}

const gchar *
gst_gl_shader_get_vertex_source (GstGLShader * shader)
{
  g_return_val_if_fail (GST_GL_IS_SHADER (shader), NULL);
  return shader->priv->vertex_src;
}

const gchar *
gst_gl_shader_get_fragment_source (GstGLShader * shader)
{
  g_return_val_if_fail (GST_GL_IS_SHADER (shader), NULL);
  return shader->priv->fragment_src;
}

static void
gst_gl_shader_init (GstGLShader * self)
{
  /* initialize sources and create program object */
  GstGLShaderPrivate *priv;

  priv = self->priv = GST_GL_SHADER_GET_PRIVATE (self);

  priv->vertex_src = NULL;
  priv->fragment_src = NULL;

  priv->fragment_handle = 0;
  priv->vertex_handle = 0;
  priv->program_handle = glCreateProgramObjectARB ();

  g_return_if_fail (priv->program_handle);

  priv->compiled = FALSE;
  priv->active = FALSE;         /* unused at the moment */

  GST_TRACE ("shader initialized %u", priv->program_handle);
}

GstGLShader *
gst_gl_shader_new (void)
{
  return g_object_new (GST_GL_TYPE_SHADER, NULL);
}

gboolean
gst_gl_shader_is_compiled (GstGLShader * shader)
{
  g_return_val_if_fail (GST_GL_IS_SHADER (shader), FALSE);

  return shader->priv->compiled;
}

gboolean
gst_gl_shader_compile (GstGLShader * shader, GError ** error)
{
  GstGLShaderPrivate *priv;

  gchar info_buffer[2048];
  GLsizei len = 0;
  GLint status = GL_FALSE;

  g_return_val_if_fail (GST_GL_IS_SHADER (shader), FALSE);

  priv = shader->priv;

  if (priv->compiled)
    return priv->compiled;

  g_return_val_if_fail (priv->program_handle, FALSE);

  if (priv->vertex_src) {
    /* create vertex object */
    const gchar *vertex_source = priv->vertex_src;
    priv->vertex_handle = glCreateShaderObjectARB (GL_VERTEX_SHADER);
    glShaderSourceARB (priv->vertex_handle, 1, &vertex_source, NULL);
    /* compile */
    glCompileShaderARB (priv->vertex_handle);
    /* check everything is ok */
    glGetObjectParameterivARB (priv->vertex_handle,
        GL_OBJECT_COMPILE_STATUS_ARB, &status);

#ifndef OPENGL_ES2
    glGetInfoLogARB (priv->vertex_handle,
        sizeof (info_buffer) - 1, &len, info_buffer);
#else
    glGetShaderInfoLog (priv->vertex_handle,
        sizeof (info_buffer) - 1, &len, info_buffer);
#endif
    info_buffer[len] = '\0';

    if (status != GL_TRUE) {
      g_set_error (error, GST_GL_SHADER_ERROR,
          GST_GL_SHADER_ERROR_COMPILE,
          "Vertex Shader compilation failed:\n%s", info_buffer);

      glDeleteObjectARB (priv->vertex_handle);
      priv->compiled = FALSE;
      return priv->compiled;
    } else if (len > 1) {
      GST_FIXME ("vertex shader info log:\n%s\n", info_buffer);
    }
    glAttachObjectARB (priv->program_handle, priv->vertex_handle);

    GST_LOG ("vertex shader attached %u", priv->vertex_handle);
  }

  if (priv->fragment_src) {
    /* create fragment object */
    const gchar *fragment_source = priv->fragment_src;
    priv->fragment_handle = glCreateShaderObjectARB (GL_FRAGMENT_SHADER_ARB);
    glShaderSourceARB (priv->fragment_handle, 1, &fragment_source, NULL);
    /* compile */
    glCompileShaderARB (priv->fragment_handle);
    /* check everything is ok */
    glGetObjectParameterivARB (priv->fragment_handle,
        GL_OBJECT_COMPILE_STATUS_ARB, &status);

#ifndef OPENGL_ES2
    glGetInfoLogARB (priv->fragment_handle,
        sizeof (info_buffer) - 1, &len, info_buffer);
#else
    glGetShaderInfoLog (priv->fragment_handle,
        sizeof (info_buffer) - 1, &len, info_buffer);
#endif
    info_buffer[len] = '\0';
    if (status != GL_TRUE) {
      g_set_error (error, GST_GL_SHADER_ERROR,
          GST_GL_SHADER_ERROR_COMPILE,
          "Fragment Shader compilation failed:\n%s", info_buffer);

      glDeleteObjectARB (priv->fragment_handle);
      priv->compiled = FALSE;
      return priv->compiled;
    } else if (len > 1) {
      GST_FIXME ("vertex shader info log:\n%s\n", info_buffer);
    }
    glAttachObjectARB (priv->program_handle, priv->fragment_handle);

    GST_LOG ("fragment shader attached %u", priv->fragment_handle);
  }

  /* if nothing failed link shaders */
  glLinkProgramARB (priv->program_handle);

#ifndef OPENGL_ES2
  glGetObjectParameterivARB (priv->program_handle, GL_LINK_STATUS, &status);
#else
  glGetProgramiv (priv->program_handle, GL_LINK_STATUS, &status);
#endif

  glGetInfoLogARB (priv->program_handle,
      sizeof (info_buffer) - 1, &len, info_buffer);
  info_buffer[len] = '\0';

  if (status != GL_TRUE) {
    g_set_error (error, GST_GL_SHADER_ERROR,
        GST_GL_SHADER_ERROR_LINK, "Shader Linking failed:\n%s", info_buffer);
    priv->compiled = FALSE;
    return priv->compiled;
  } else if (len > 1) {
    GST_FIXME ("shader link log:\n%s\n", info_buffer);
  }
  /* success! */
  priv->compiled = TRUE;
  g_object_notify (G_OBJECT (shader), "compiled");

  return priv->compiled;
}

void
gst_gl_shader_release (GstGLShader * shader)
{
  GstGLShaderPrivate *priv;
  /* GLint status; */
  /* GLenum err = 0; */

  g_return_if_fail (GST_GL_IS_SHADER (shader));

  priv = shader->priv;

  g_return_if_fail (priv->program_handle);

  if (!priv->compiled)
    return;

  if (priv->vertex_handle) {    /* not needed but nvidia doesn't care to respect the spec */
    GST_TRACE ("finalizing vertex shader %u", priv->vertex_handle);

#ifndef OPENGL_ES2
    glDeleteObjectARB (priv->vertex_handle);
#else
    glDeleteShader (priv->vertex_handle);
#endif

    /* err = glGetError (); */
    /* GST_WARNING ("error: 0x%x", err); */
    /* glGetObjectParameterivARB(priv->vertex_handle, GL_OBJECT_DELETE_STATUS_ARB, &status); */
    /* GST_INFO ("vertex deletion status:%s", status == GL_TRUE ? "true" : "false" ); */
  }

  if (priv->fragment_handle) {
    GST_TRACE ("finalizing fragment shader %u", priv->fragment_handle);

#ifndef OPENGL_ES2
    glDeleteObjectARB (priv->fragment_handle);
#else
    glDeleteShader (priv->fragment_handle);
#endif

    /* err = glGetError (); */
    /* GST_WARNING ("error: 0x%x", err); */
    /* glGetObjectParameterivARB(priv->fragment_handle, GL_OBJECT_DELETE_STATUS_ARB, &status); */
    /* GST_INFO ("fragment deletion status:%s", status == GL_TRUE ? "true" : "false" ); */
  }

  if (priv->vertex_handle)
    glDetachObjectARB (priv->program_handle, priv->vertex_handle);
  if (priv->fragment_handle)
    glDetachObjectARB (priv->program_handle, priv->fragment_handle);

  priv->compiled = FALSE;
  g_object_notify (G_OBJECT (shader), "compiled");
}

void
gst_gl_shader_use (GstGLShader * shader)
{
  GstGLShaderPrivate *priv;

  if (!shader) {
    glUseProgramObjectARB (0);
    return;
  }

  priv = shader->priv;

  g_return_if_fail (priv->program_handle);

  glUseProgramObjectARB (priv->program_handle);

  return;
}

gboolean
gst_gl_shader_compile_and_check (GstGLShader * shader,
    const gchar * source, GstGLShaderSourceType type)
{
  gboolean is_compiled = FALSE;

  g_object_get (G_OBJECT (shader), "compiled", &is_compiled, NULL);

  if (!is_compiled) {
    GError *error = NULL;

    switch (type) {
      case GST_GL_SHADER_FRAGMENT_SOURCE:
        gst_gl_shader_set_fragment_source (shader, source);
        break;
      case GST_GL_SHADER_VERTEX_SOURCE:
        gst_gl_shader_set_vertex_source (shader, source);
        break;
      default:
        g_assert_not_reached ();
        break;
    }

    gst_gl_shader_compile (shader, &error);
    if (error) {
      GST_WARNING ("%s", error->message);
      g_error_free (error);
      error = NULL;
      gst_gl_shader_use (NULL);
      return FALSE;
    }
  }
  return TRUE;
}

void
gst_gl_shader_set_uniform_1f (GstGLShader * shader, const gchar * name,
    gfloat value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform1fARB (location, value);
}

void
gst_gl_shader_set_uniform_1fv (GstGLShader * shader, const gchar * name,
    guint count, gfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform1fvARB (location, count, value);
}

void
gst_gl_shader_set_uniform_1i (GstGLShader * shader, const gchar * name,
    gint value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform1iARB (location, value);
}

void
gst_gl_shader_set_uniform_1iv (GstGLShader * shader, const gchar * name,
    guint count, gint * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform1ivARB (location, count, value);
}

void
gst_gl_shader_set_uniform_2f (GstGLShader * shader, const gchar * name,
    gfloat value0, gfloat value1)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform2fARB (location, value0, value1);
}

void
gst_gl_shader_set_uniform_2fv (GstGLShader * shader, const gchar * name,
    guint count, gfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform2fvARB (location, count, value);
}

void
gst_gl_shader_set_uniform_2i (GstGLShader * shader, const gchar * name,
    gint v0, gint v1)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform2iARB (location, v0, v1);
}

void
gst_gl_shader_set_uniform_2iv (GstGLShader * shader, const gchar * name,
    guint count, gint * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform2ivARB (location, count, value);
}

void
gst_gl_shader_set_uniform_3f (GstGLShader * shader, const gchar * name,
    gfloat v0, gfloat v1, gfloat v2)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform3fARB (location, v0, v1, v2);
}

void
gst_gl_shader_set_uniform_3fv (GstGLShader * shader, const gchar * name,
    guint count, gfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform3fvARB (location, count, value);
}

void
gst_gl_shader_set_uniform_3i (GstGLShader * shader, const gchar * name,
    gint v0, gint v1, gint v2)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform3iARB (location, v0, v1, v2);
}

void
gst_gl_shader_set_uniform_3iv (GstGLShader * shader, const gchar * name,
    guint count, gint * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform3ivARB (location, count, value);
}

void
gst_gl_shader_set_uniform_4f (GstGLShader * shader, const gchar * name,
    gfloat v0, gfloat v1, gfloat v2, gfloat v3)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform4fARB (location, v0, v1, v2, v3);
}

void
gst_gl_shader_set_uniform_4fv (GstGLShader * shader, const gchar * name,
    guint count, gfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform4fvARB (location, count, value);
}

void
gst_gl_shader_set_uniform_4i (GstGLShader * shader, const gchar * name,
    gint v0, gint v1, gint v2, gint v3)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform4iARB (location, v0, v1, v2, v3);
}

void
gst_gl_shader_set_uniform_4iv (GstGLShader * shader, const gchar * name,
    guint count, gint * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniform4ivARB (location, count, value);
}

void
gst_gl_shader_set_uniform_matrix_2fv (GstGLShader * shader, const gchar * name,
    GLsizei count, GLboolean transpose, const GLfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniformMatrix2fvARB (location, count, transpose, value);
}

#ifndef OPENGL_ES2
void
gst_gl_shader_set_uniform_matrix_2x3fv (GstGLShader * shader,
    const gchar * name, GLsizei count, GLboolean transpose,
    const GLfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniformMatrix2x3fv (location, count, transpose, value);
}

void
gst_gl_shader_set_uniform_matrix_2x4fv (GstGLShader * shader,
    const gchar * name, GLsizei count, GLboolean transpose,
    const GLfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniformMatrix2x4fv (location, count, transpose, value);
}
#endif

void
gst_gl_shader_set_uniform_matrix_3fv (GstGLShader * shader, const gchar * name,
    GLsizei count, GLboolean transpose, const GLfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniformMatrix3fvARB (location, count, transpose, value);
}

#ifndef OPENGL_ES2
void
gst_gl_shader_set_uniform_matrix_3x2fv (GstGLShader * shader,
    const gchar * name, GLsizei count, GLboolean transpose,
    const GLfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniformMatrix3x2fv (location, count, transpose, value);
}

void
gst_gl_shader_set_uniform_matrix_3x4fv (GstGLShader * shader,
    const gchar * name, GLsizei count, GLboolean transpose,
    const GLfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniformMatrix3x4fv (location, count, transpose, value);
}
#endif

void
gst_gl_shader_set_uniform_matrix_4fv (GstGLShader * shader, const gchar * name,
    GLsizei count, GLboolean transpose, const GLfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniformMatrix4fvARB (location, count, transpose, value);
}

#ifndef OPENGL_ES2
void
gst_gl_shader_set_uniform_matrix_4x2fv (GstGLShader * shader,
    const gchar * name, GLsizei count, GLboolean transpose,
    const GLfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniformMatrix4x2fv (location, count, transpose, value);
}

void
gst_gl_shader_set_uniform_matrix_4x3fv (GstGLShader * shader,
    const gchar * name, GLsizei count, GLboolean transpose,
    const GLfloat * value)
{
  GstGLShaderPrivate *priv;
  GLint location = -1;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  location = glGetUniformLocationARB (priv->program_handle, name);

  glUniformMatrix4x3fv (location, count, transpose, value);
}
#endif

GLint
gst_gl_shader_get_attribute_location (GstGLShader * shader, const gchar * name)
{
  GstGLShaderPrivate *priv;

  priv = shader->priv;

  g_return_val_if_fail (priv->program_handle != 0, 0);

  return glGetAttribLocationARB (priv->program_handle, name);
}

void
gst_gl_shader_bind_attribute_location (GstGLShader * shader, GLuint index,
    const gchar * name)
{
  GstGLShaderPrivate *priv;

  priv = shader->priv;

  g_return_if_fail (priv->program_handle != 0);

  glBindAttribLocationARB (priv->program_handle, index, name);
}

GQuark
gst_gl_shader_error_quark (void)
{
  return g_quark_from_static_string ("gst-gl-shader-error");
}
